# Copyright 2022 Touca, Inc. Subject to Apache-2.0 License.

import logging
from argparse import ArgumentParser
from pathlib import Path

import py7zr
from rich.progress import Progress
from touca.cli._common import Operation, ResultsTree
from touca._options import find_home_path

logger = logging.getLogger("touca.cli.zip")


def _compress_batch(binary_files, zip_file, update):
    if zip_file.exists():
        logger.debug(f"Compressed file {zip_file} already exists")
        update(len(binary_files))
        return True
    logger.debug(f"Creating {zip_file}")
    try:
        with py7zr.SevenZipFile(zip_file, "w") as archive:
            for binary_file in binary_files:
                archive.write(binary_file, arcname=binary_file.name)
                update(1)
    except py7zr.ArchiveError:
        logger.warning(f"Failed to archive {zip_file}")
        return False
    return True


class Zip(Operation):
    name = "zip"
    help = "Compress binary archives"

    def __init__(self, options: dict):
        self.__options = options

    @classmethod
    def parser(self, parser: ArgumentParser):
        home_dir = find_home_path()
        parser.add_argument(
            "src",
            nargs="?",
            default=home_dir.joinpath("results"),
            help=f"Directory with binary files. Defaults to {home_dir.joinpath('results')}",
        )
        parser.add_argument(
            "out",
            nargs="?",
            default=home_dir.joinpath("zip"),
            help=f"Directory to store compressed files. Defaults to {home_dir.joinpath('zip')}",
        )

    def run(self):
        src_dir = Path(self.__options.get("src")).expanduser().resolve()
        out_dir = Path(self.__options.get("out")).expanduser().resolve()

        results_tree = ResultsTree(src_dir)
        if results_tree.is_empty():
            logger.error(f"Did not find any binary file in {src_dir}")
            return False

        for suite_name, versions in results_tree.suites.items():
            zip_dir = out_dir.joinpath(suite_name)
            if not zip_dir.exists():
                zip_dir.mkdir(parents=True, exist_ok=True)
            for version_name, binary_files in versions.items():
                zip_file = zip_dir.joinpath(version_name + ".7z")
                with Progress() as progress:
                    task_batch = progress.add_task(
                        f"[magenta]{suite_name}/{version_name}[/magenta]",
                        total=len(binary_files),
                    )
                    update = lambda x: progress.update(task_batch, advance=x)
                    if not _compress_batch(binary_files, zip_file, update=update):
                        logger.error(f"failed to compress {src_dir}")
                        return False
        return True



# Copyright 2022 Touca, Inc. Subject to Apache-2.0 License.

import sys
from argparse import ArgumentParser
from pathlib import Path

from touca._options import find_home_path
from touca.cli._common import Operation, invalid_subcommand


def user_plugins():
    import importlib
    import inspect
    import sys

    plugins_dir = Path(find_home_path(), "plugins")
    modules = [p.absolute() for p in plugins_dir.glob("*") if p.is_file()]
    for module in modules:
        syspath = Path(module.parent).absolute()
        sys.path.append(f"{syspath}/")
        mod = importlib.import_module(module.stem)
        for (_, member) in inspect.getmembers(mod):
            if not inspect.isclass(member):
                continue
            tree = inspect.getclasstree(inspect.getmro(member), unique=True)
            tmp = tree[1][-1][-1]
            if type(tmp) is list:
                yield member
        sys.path.remove(f"{syspath}/")


class Plugin(Operation):
    name = "plugin"
    help = "Install and manage custom CLI plugins"

    @classmethod
    def parser(self, parser: ArgumentParser):
        parsers = parser.add_subparsers(dest="subcommand")
        parsers.add_parser(
            "ls",
            description="List available plugins",
            help="List available plugins",
        )
        parsers_add = parsers.add_parser(
            "add",
            description="Install a plugin",
            help="Install a plugin",
        )
        parsers_add.add_argument("name", help="name of the plugin")
        parsers_remove = parsers.add_parser(
            "rm",
            description="Uninstall a plugin",
            help="Uninstall a plugin",
        )
        parsers_remove.add_argument("name", help="name of the plugin")
        parsers.add_parser(
            "new", description="Create a new plugin", help="Create a new plugin"
        )

    def __init__(self, options: dict):
        self.__options = options

    def _command_list(self):
        from touca._printer import print_table

        plugins = list(user_plugins())
        if not plugins:
            print("No user-defined plugins are registered.")
            return True
        table_header = ["", "Name", "Description"]
        table_body = [
            [f"{idx + 1}", member.name, member.help]
            for idx, member in enumerate(plugins)
        ]
        print_table(table_header, table_body)
        return True

    def _command_add(self):
        from shutil import copyfile

        plugin_arg: str = self.__options.get("name")
        plugin_name = plugin_arg
        plugin_path_dst = Path(find_home_path(), "plugins", plugin_name).with_suffix(
            ".py"
        )
        plugin_path_dst.parent.mkdir(exist_ok=True)
        if plugin_path_dst.exists():
            print(f'plugin "{plugin_name}" is already installed', file=sys.stderr)
            return False
        plugin_path_src = Path.cwd().joinpath(plugin_name).with_suffix(".py")
        if plugin_path_src.exists():
            copyfile(plugin_path_src, plugin_path_dst)
            return True
        print(f'did not find a plugin at "{plugin_arg}"', file=sys.stderr)
        return False

    def _command_remove(self):
        plugin_name = self.__options.get("name")
        plugin_path_dst = Path(find_home_path(), "plugins", plugin_name).with_suffix(
            ".py"
        )
        if not plugin_path_dst.exists():
            print(f'plugin "{plugin_name}" is missing', file=sys.stderr)
            return False
        Path.unlink(plugin_path_dst)
        return True

    def _command_template(self):
        content = """
from argparse import ArgumentParser

from touca.cli._common import Operation


class Example(Operation):
    name = "example"
    help = "Example"

    @classmethod
    def parser(self, parser: ArgumentParser):
        parser.add_argument("args", nargs="+", help="any problem")

    def run(self):
        print("Example!")
        return True
"""
        dir = Path.cwd()
        file = "example.py"
        dir.joinpath(file).write_text(content)
        print(
            f'Created plugin "{file}" in "{dir}" for you to implement.\n'
            f'Run "touca plugin add {file}" when you are ready to register it.'
        )
        return True

    def run(self):
        commands = {
            "add": self._command_add,
            "ls": self._command_list,
            "rm": self._command_remove,
            "new": self._command_template,
        }
        command = self.__options.get("subcommand")
        if not command:
            return invalid_subcommand(Plugin)
        if command in commands:
            return commands.get(command)()
        return False



# Copyright 2022 Touca, Inc. Subject to Apache-2.0 License.

import logging
from argparse import ArgumentParser
from pathlib import Path

from touca._options import find_home_path
from touca._printer import print_results_tree
from touca.cli._common import Operation, ResultsTree, invalid_subcommand

logger = logging.Logger("touca.cli.results")


class Results(Operation):
    name = "results"
    help = "Show suite results"

    def __init__(self, options: dict):
        self.__options = options

    @classmethod
    def parser(self, parser: ArgumentParser):
        parsers = parser.add_subparsers(dest="subcommand")
        parser_ls = parsers.add_parser("ls", help="list local touca archive files")
        parser_rm = parsers.add_parser("rm", help="remove local touca archive files")
        results_dir = find_home_path().joinpath("results")
        for parser in (parser_ls, parser_rm):
            parser.add_argument(
                "--src",
                dest="src_dir",
                help=f"path to test results directory. defaults to {results_dir}.",
                default=results_dir,
            )
            parser.add_argument(
                "--filter",
                help="limit results to a given suite or version. value should be in form of suite[/version].",
                default=None,
            )
        parser_rm.add_argument(
            "--dry-run",
            action="store_true",
            dest="dry_run",
            help="Check what your command would do when run without this option",
        )

    def _command_ls(self):
        filter = self.__options.get("filter", None)
        src_dir = self.__options.get("src_dir")
        results_tree = ResultsTree(src_dir, filter)
        print_results_tree(results_tree.suites)
        return True

    def _command_rm(self):
        from rich.progress import Progress
        from shutil import rmtree

        filter = self.__options.get("filter", None)
        src_dir: Path = self.__options.get("src_dir")
        dry_run = self.__options.get("dry_run", False)
        results_tree = ResultsTree(src_dir, filter)

        if dry_run:
            for versions in results_tree.suites.values():
                for binary_files in versions.values():
                    for binary_file in binary_files:
                        logger.info(f"will remove {binary_file}")
            return True

        with Progress() as progress:
            for suite, versions in results_tree.suites.items():
                for version, binary_files in versions.items():
                    task_name = f"[magenta]{suite}/{version}[/magenta]"
                    task_batch = progress.add_task(task_name, total=len(binary_files))
                    for binary_file in binary_files:
                        logger.debug(f"removing {binary_file}")
                        binary_file.unlink()
                        progress.update(task_batch, advance=1)
                    rmtree(src_dir.joinpath(suite, version))
                rmtree(src_dir.joinpath(suite))
        return True

    def run(self):
        commands = {
            "ls": self._command_ls,
            "rm": self._command_rm,
        }
        command = self.__options.get("subcommand")
        if not command:
            return invalid_subcommand(Results)
        if command in commands:
            return commands.get(command)()
        return False



# Copyright 2022 Touca, Inc. Subject to Apache-2.0 License.

import argparse
import logging
import sys
from typing import List

from touca import __version__
from touca._options import find_home_path
from touca._printer import Printer
from touca.cli.check import Check
from touca.cli._common import Operation
from touca.cli._config import Config
from touca.cli._execute import Execute
from touca.cli._merge import Merge
from touca.cli._plugin import Plugin, user_plugins
from touca.cli._post import Post
from touca.cli._profile import Profile
from touca.cli._results import Results
from touca.cli._run import Run
from touca.cli._unzip import Unzip
from touca.cli._update import Update
from touca.cli._zip import Zip
from touca.cli.server import Server
from rich.logging import RichHandler


def _find_latest_pypi_version():
    from json import loads
    from urllib.request import urlopen

    with urlopen("https://pypi.org/pypi/touca/json") as response:
        data = loads(response.read())
        return data["info"]["version"]


def _warn_outdated_version():
    from packaging import version

    latest_version = _find_latest_pypi_version()
    if version.parse(latest_version) <= version.parse(__version__):
        return
    fmt = (
        "You are using touca version {}; however, version {} is available."
        + "\nConsider upgrading by running 'pip install --upgrade touca'."
    )
    Printer.print_warning(fmt, __version__, latest_version)


class Help(Operation):
    name = "help"
    help = "Shows this help message"

    def __init__(self, options):
        self._options = options

    @classmethod
    def parser(self, parser: argparse.ArgumentParser):
        parser.add_argument(
            "subcommand", help="subcommand to get help about", nargs="?"
        )

    def run(self, parser: argparse.ArgumentParser, subcommands: List[Operation]):
        subcommand_name = self._options.get("subcommand")
        subcommand = next((x for x in subcommands if x.name == subcommand_name), None)
        if not subcommand:
            parser.print_help()
            return False
        subcommand_parser = argparse.ArgumentParser(
            prog=f"touca {subcommand_name}",
            add_help=False,
            epilog="See https://touca.io/docs/cli for more information.",
        )
        subcommand.parser(subcommand_parser)
        subcommand_parser.print_help()
        return False


class Version(Operation):
    name = "version"
    help = "Check your Touca CLI version"

    def __init__(self, options):
        pass

    def run(self):
        print(f"v{__version__}")
        return True


def main(args=None):
    subcommands = [
        Check,
        Config,
        Help,
        Merge,
        Plugin,
        Post,
        Profile,
        Results,
        Run,
        Server,
        Execute,
        Unzip,
        Update,
        Version,
        Zip,
        *user_plugins(),
    ]
    parser = argparse.ArgumentParser(
        prog="touca",
        add_help=False,
        description="Work seamlessly with Touca from the command line.",
        epilog="See https://touca.io/docs/cli for more information.",
    )
    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version=f"%(prog)s v{__version__}",
        help=argparse.SUPPRESS,
    )
    parsers = parser.add_subparsers(dest="command")
    for operation in subcommands:
        subparser = parsers.add_parser(
            name=operation.name,
            prog=f"touca {operation.name}",
            description=operation.help,
            help=operation.help,
            add_help=True,
        )
        if callable(getattr(operation, "parser", None)):
            operation.parser(subparser)
    parsed, remaining = parser.parse_known_args(sys.argv[1:] if args is None else args)
    options = vars(parsed)

    command = next((x for x in subcommands if x.name == options.get("command")), None)
    if (
        not command
        or command is Help
        or any(arg in remaining for arg in ["-h", "--help"])
    ):
        return Help(options).run(parser, subcommands)
    operation = command(options)

    home_dir = find_home_path()
    home_dir.mkdir(parents=True, exist_ok=True)

    logging.basicConfig(
        format="%(message)s",
        handlers=[RichHandler(show_path=False, show_time=False)],
        level=logging.INFO,
    )

    if not operation.run():
        return True

    _warn_outdated_version()
    return False


if __name__ == "__main__":
    sys.exit(main())



# Copyright 2022 Touca, Inc. Subject to Apache-2.0 License.

import logging
from argparse import ArgumentParser
from pathlib import Path
from typing import List
import sys
import touca
from touca.cli._common import Operation
from touca._runner import run_workflows, _parse_cli_options, _Workflow

logger = logging.getLogger("touca.cli.check")


class Check(Operation):
    name = "check"
    help = "Submit the test output generated by other applications to the Touca Server"

    @classmethod
    def parser(self, parser: ArgumentParser):
        parser.add_argument(
            "--suite",
            help="name of the suite to associate with this output",
            required=True,
        )
        parser.add_argument(
            "--testcase",
            help="name of the testcase to associate with this output",
            required=False,
        )
        if sys.stdin.isatty():
            parser.add_argument(
                "src",
                help="path to file or directory to submit",
            )

    def __init__(self, options: dict):
        self.__options = options

    def _slugify(self, file: Path):
        return (
            str(file.absolute().relative_to(Path.cwd()))
            .replace(".", "_")
            .replace("/", "_")
            .replace("-", "_")
        )

    def _submit_stdin(self):
        def _submit(testcase: str):
            touca.check("output", sys.stdin.read())

        workflow = _Workflow(_submit, self.__options.get("suite"))
        testcase = self.__options.get("testcase")
        workflow.testcases = [testcase] if testcase else ["stdout"]
        run_workflows(_parse_cli_options([]), [workflow])
        return True

    def _submit_file(self, file: Path):
        content = file.read_text()

        def _submit(testcase: str):
            touca.check("output", content)

        testcase = self.__options.get("testcase")
        workflow = _Workflow(_submit, self.__options.get("suite"))
        workflow.testcases = [testcase] if testcase else [self._slugify(file)]
        run_workflows(_parse_cli_options([]), [workflow])
        return True

    def _get_file_content(self, file: Path):
        try:
            return file.read_text()
        except:
            from hashlib import sha256

            return sha256(file.read_bytes()).hexdigest()

    def _submit_directory(self, directory: Path):
        files = {
            self._slugify(file): file for file in directory.glob("*") if file.is_file()
        }

        def _submit(testcase: str):
            if not self.__options.get("testcase"):
                content = self._get_file_content(files.get(testcase))
                touca.check("output", content)
                return
            for slug, file in files.items():
                touca.check(slug, self._get_file_content(file))

        testcase = self.__options.get("testcase")
        workflow = _Workflow(_submit, self.__options.get("suite"))
        workflow.testcases = [testcase] if testcase else files.keys()
        run_workflows(_parse_cli_options([]), [workflow])
        return False

    def run(self) -> bool:
        if not sys.stdin.isatty():
            return self._submit_stdin()
        src = Path(self.__options.get("src"))
        if src.is_file():
            return self._submit_file(src)
        if src.is_dir():
            return self._submit_directory(src)
        logger.error("specified path is neither a directory nor a file")
        return False



# Copyright 2022 Touca, Inc. Subject to Apache-2.0 License.

import sys
from argparse import ArgumentParser
from pathlib import Path

from touca._options import config_file_parse
from touca._runner import Workflow, prepare_parser, run_workflows
from touca.cli._common import Operation
import logging


def is_test_module(module: str):
    with open(module, "rt") as file:
        content = file.read()
        return any(x in content for x in ["@touca.workflow", "@touca.Workflow"])


def find_test_modules(testdir: str):
    return [p for p in Path(testdir).glob("**/*.py") if is_test_module(p)]


def load_workflows(modules: list):
    import importlib
    import sys

    for module in modules:
        relpath = Path(module).relative_to(Path.cwd())
        syspath = Path(relpath.parent).absolute()
        sys.path.append(f"{syspath}/")
        importlib.import_module(relpath.stem)
        sys.path.remove(f"{syspath}/")


class Execute(Operation):
    name = "test"
    help = "Run your Touca tests"

    @classmethod
    def parser(cls, parser: ArgumentParser):
        parser.add_argument(
            "--testdir",
            nargs=1,
            help="path to regression tests directory",
        )
        prepare_parser(parser)

    def __init__(self, options: dict):
        self.__options = {k: v for k, v in options.items() if v is not None}

    def _find_arguments(self):
        args = self.__options
        flags = [
            "save-as-binary",
            "save-as-json",
            "offline",
            "overwrite",
            "colored-output",
        ]
        config_content = config_file_parse()
        if config_content:
            for key in config_content.options("settings"):
                func = config_content.getboolean if key in flags else config_content.get
                args.update({key: func("settings", key)})
        return args

    def run(self):
        logging.disable(logging.CRITICAL)
        dir_test = Path(self.__options.get("testdir", [Path.cwd()])[0]).resolve()
        args = self._find_arguments()
        modules = find_test_modules(dir_test)
        load_workflows(modules)
        try:
            run_workflows(args, Workflow._workflows)
        except Exception as err:
            print(f"test failed: {err}", file=sys.stderr)
            return False
        return True



# Copyright 2022 Touca, Inc. Subject to Apache-2.0 License.

import json
import logging
from argparse import ArgumentParser
from distutils.version import LooseVersion
from pathlib import Path
from typing import Dict, List

from rich.progress import Progress
from touca._transport import Transport
from touca.cli._common import Operation, ResultsTree
from touca._options import find_home_path

logger = logging.getLogger("touca.cli.post")


class Post(Operation):
    name = "post"
    help = "Submit binary archives to a Touca server"

    def __init__(self, options: dict):
        self.__options = options
        self._transport: Transport = None

    @classmethod
    def parser(self, parser: ArgumentParser):
        results_dir = find_home_path().joinpath("results")
        parser.add_argument(
            "src",
            help=f"path to directory with binary files. defaults to {results_dir}",
            nargs="?",
            default=results_dir,
        )
        group_credentials = parser.add_argument_group(
            "Credentials",
            'Server API Key and URL. Not required when specified in the active configuration profile. Ignored when "--dry-run" is specified.',
        )
        group_credentials.add_argument(
            "--api-key", dest="api-key", help="Touca API Key", required=False
        )
        group_credentials.add_argument(
            "--api-url", dest="api-url", help="Touca API URL", required=False
        )
        group_misc = parser.add_argument_group("Miscellaneous")
        group_misc.add_argument(
            "--dry-run",
            action="store_true",
            dest="dry-run",
            help="Check what your command would do when run without this option",
        )

    def run(self):
        self._update_options()

        if not self._setup_transport():
            return False

        src = Path(self.__options.get("src")).expanduser().resolve()
        results_tree = ResultsTree(src)
        if results_tree.is_empty():
            logger.error(f"Did not find any binary file in {src}")
            return False

        errors = self._process(results_tree)
        self._print_errors(errors)
        return not errors

    def _update_options(self):
        from touca._options import _apply_config_file, update_options

        options = {
            k: self.__options.get(k)
            for k in ["api-key", "api-url", "dry-run"]
            if self.__options.get(k) is not None
        }
        options.update({"suite": "", "version": ""})

        _apply_config_file(options)
        if "team" in options and "api-url" not in options:
            options["api-url"] = "https://api.touca.io"

        try:
            update_options(options, options)
        except ValueError as err:
            logger.error(err)
            return False
        self.__options.update(options)

    def _setup_transport(self):
        if self.__options.get("dry-run"):
            logger.warning("Running in dry-run mode")
            return True
        try:
            self._transport = Transport(
                {k: self.__options.get(k) for k in ["api-key", "api-url"]}
            )
            self._transport.authenticate()
        except ValueError as err:
            logger.error(err)
            return False
        return True

    def _process(self, results_tree: ResultsTree):
        errors = {}
        with Progress() as progress:
            for suite_name, batches in results_tree.suites.items():
                for batch_name, binary_files in batches.items():
                    task_name = f"[magenta]{suite_name}/{batch_name}[/magenta]"
                    task_batch = progress.add_task(task_name, total=len(binary_files))
                    for binary_file in binary_files:
                        error = self._post(binary_file)
                        logger.debug(f"processed {binary_file}")
                        progress.update(task_batch, advance=1)
                        if not error:
                            continue
                        if error not in errors:
                            errors[error] = []
                        errors[error].append(binary_file)
        return errors

    def _post(self, binary_file: Path):
        if not self._transport:
            return None
        content = binary_file.read_bytes()
        response = self._transport._send_request(
            method="POST",
            path=f"/client/submit",
            body=content,
            content_type="application/octet-stream",
        )
        if response.status != 200 and response.data:
            body = json.loads(response.data.decode("utf-8"))
            return body["errors"][0]

    def _print_errors(self, errors: Dict[str, List[str]]):
        if not errors:
            return
        logger.error(f"Failed to post some binary files")
        for error, binary_files in errors.items():
            binary_files.sort()
            logger.error(f" {error}")
            for binary_file in binary_files:
                logger.error(f"  {binary_file}")



# Copyright 2022 Touca, Inc. Subject to Apache-2.0 License.

import logging
from argparse import ArgumentParser
from pathlib import Path

from colorama import Style
from touca.cli._common import Operation, invalid_subcommand

logger = logging.getLogger("touca.cli.server")


class Server(Operation):
    name = "server"
    help = "Install and manage your Touca Server"

    @classmethod
    def parser(self, parser: ArgumentParser):
        parsers = parser.add_subparsers(dest="subcommand")
        parser_install = parsers.add_parser(
            "install",
            description="Install touca server",
            help="Install and run touca server.",
        )
        parser_install.add_argument(
            "--dev",
            action="store_true",
            dest="dev",
            help="Install for development.",
        )
        group_misc = parser.add_argument_group("Miscellaneous")
        group_misc.add_argument(
            "--dry-run",
            action="store_true",
            dest="dry-run",
            help="Check what your command would do when run without this option",
        )

    def __init__(self, options: dict):
        self.__options = options

    def _ask(self, question: str, default: str = ""):
        output = input(f"{question}\n> ")
        return output if output else default

    def _ask_name(self):
        print("Hi, Thank you for trying Touca!")
        output = self._ask("What is your first name?", "stranger")
        print(f"Nice to meet you, {output}!")
        return output

    def _ask_install_dir(self):
        default = Path.home().joinpath(".touca/server")
        output = self._ask(
            "Where should we install Touca? "
            f'{Style.DIM}(default is "{default}"){Style.RESET_ALL}',
            default,
        )
        print(f"Installing into {output}")
        return output

    def _command_install(self):
        user_name = self._ask_name()
        install_dir = self._ask_install_dir()
        return False

    def run(self) -> bool:
        print(self.__options)
        commands = {"install": self._command_install}
        command = self.__options.get("subcommand")
        if not command:
            return invalid_subcommand(Server)
        if command in commands:
            return commands.get(command)()
        return False



# Copyright 2022 Touca, Inc. Subject to Apache-2.0 License.

from abc import ABC, abstractmethod
from pathlib import Path
from typing import Dict, List


class Operation(ABC):
    @abstractmethod
    def run(self) -> bool:
        pass


class ResultsTree:
    suites: Dict[str, Dict[str, List[Path]]] = {}

    def __init__(self, src: Path, filter: str = None):
        self._filters = filter.split("/") if filter else []
        self._filters.extend([None, None])
        if not src.exists():
            return
        if src.is_file():
            self._process(src)
            return
        for binary_file in src.rglob("*.bin"):
            self._process(binary_file)

    def _process(self, binary_file: Path):
        testcase_dir = binary_file.parent
        version_dir = testcase_dir.parent
        version_name = version_dir.name
        suite_dir = version_dir.parent
        suite_name = suite_dir.name

        if self._filters[0] is not None and self._filters[0] != suite_name:
            return
        if self._filters[1] is not None and self._filters[1] != version_name:
            return

        if suite_name not in self.suites:
            self.suites[suite_name] = {}
        if version_name not in self.suites[suite_name]:
            self.suites[suite_name][version_name] = []
        self.suites[suite_name][version_name].append(binary_file)

    def is_empty(self):
        return len(self) == 0

    def __len__(self):
        return sum(sum(len(x) for x in bs.values()) for bs in self.suites.values())


def invalid_subcommand(cls):
    import sys
    from argparse import ArgumentParser

    parser = ArgumentParser(prog=f"touca {cls.name}", description=cls.help)
    cls.parser(parser)
    parser.print_help(file=sys.stderr)
    return False
